# Copyright (c) 2006-2008 L. C. Rees
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice, 
#       this list of conditions and the following disclaimer.
#    
#    2. Redistributions in binary form must reproduce the above copyright 
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#    3. Neither the name of Django nor the names of its contributors may be used
#       to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''RESTful WSGI URL dispatcher.'''

import re
import time
import copy
import random
import threading
from fnmatch import translate

__author__ = 'L.C. Rees (lcrees@gmail.com)'
__revision__ = '0.7'
__all__ = ['URLRelay', 'url', 'register']

def synchronized(func):
    '''Decorator to lock and unlock a method (Phillip J. Eby).

    @param func Method to decorate
    '''
    def wrapper(self, *__args, **__kw):
        self._lock.acquire()
        try:
            return func(self, *__args, **__kw)
        finally:
            self._lock.release()
    wrapper.__name__ = func.__name__
    wrapper.__dict__ = func.__dict__
    wrapper.__doc__ = func.__doc__
    return wrapper

def _handler(environ, start_response):
    start_response('404 Not Found', [('content-type', 'text/plain')])
    return ['Requested URL was not found on this server.']

def pattern_compile(pattern, pattern_type):
    '''Compile pattern.'''
    # Convert glob expression to regex
    if pattern_type == 'glob': pattern = translate(pattern)
    return re.compile(pattern)


class _Registry(object):

    '''Maintains order of URL preference while updating the central URL path
    registry.'''

    _register = list()

    def __iter__(self):
        '''Iterator for registry.'''
        return iter(self._register)   

    def add(self, pattern, mapping):
        '''Add tuple to registry.

        @param pattern URL pattern
        @param mapping WSGI application mapping
        '''        
        self._register.append((pattern, mapping))

    def get(self):
        '''Returns current registry.'''
        return tuple(self._register)


# URL registry
_reg = _Registry()

def register(pattern, application, method=None):
    '''Registers a pattern, application, and optional HTTP method.

    @param pattern URL pattern
    @param application WSGI application
    @param method HTTP method (default: None)
    '''
    if method is None:
        _reg.add(pattern, application)
    # Handle URL/method combinations
    else:
        # Update any existing registry entry
        for entry in _reg:
            if entry[0] == pattern:
                entry[1][method] = application
                return None
        # Add new registry entry
        _reg.add(pattern, {method:application})

def url(pattern, method=None):
    '''Decorator for registering a path pattern /application pair.

    @param pattern Regex pattern for path
    @param method HTTP method (default: None)
    '''
    def decorator(application):
        register(pattern, application, method)
        return application
    return decorator


class lazy(object):

    '''Lazily assign attributes on an instance upon first use.'''

    def __init__(self, method):
        self.method = method
        self.name = method.__name__

    def __get__(self, instance, cls):
        if instance is None: return self
        value = self.method(instance)
        setattr(instance, self.name, value)
        return value


class MemoryCache(object):

    '''Base Cache class.'''    
    
    def __init__(self, **kw):
        # Set timeout
        timeout = kw.get('timeout', 300)
        try:
            timeout = int(timeout)
        except (ValueError, TypeError):
            timeout = 300
        self.timeout = timeout
        # Get random seed
        random.seed()
        self._cache = dict()
        # Set max entries
        max_entries = kw.get('max_entries', 300)
        try:
            self._max_entries = int(max_entries)
        except (ValueError, TypeError):
            self._max_entries = 300
        # Set maximum number of items to cull if over max
        self._maxcull = kw.get('maxcull', 10)
        self._lock = threading.Condition()

    def __getitem__(self, key):
        '''Fetch a given key from the cache.'''
        return self.get(key)

    def __setitem__(self, key, value):
        '''Set a value in the cache. '''
        self.set(key, value)

    def __delitem__(self, key):
        '''Delete a key from the cache.'''
        self.delete(key) 

    def __contains__(self, key):
        '''Tell if a given key is in the cache.'''
        return self.get(key) is not None   

    def get(self, key, default=None):
        '''Fetch a given key from the cache.  If the key does not exist, return
        default, which itself defaults to None.

        @param key Keyword of item in cache.
        @param default Default value (default: None)
        '''
        values = self._cache.get(key)
        if values is None: 
            value = default
        elif values[0] < time.time():
            self.delete(key)
            value = default
        else:
            value = values[1]
        return copy.deepcopy(value)

    def set(self, key, value):
        '''Set a value in the cache.

        @param key Keyword of item in cache.
        @param value Value to be inserted in cache.
        '''
        # Cull timed out values if over max # of entries
        if len(self._cache) >= self._max_entries: self._cull()
        # Set value and timeout in cache
        self._cache[key] = (time.time() + self.timeout, value)

    def delete(self, key):
        '''Delete a key from the cache, failing silently.

        @param key Keyword of item in cache.
        '''
        try:
            del self._cache[key]
        except KeyError: 
            pass

    def keys(self):
        '''Returns a list of keys in the cache.'''
        return self._cache.keys()

    def _cull(self):
        '''Remove items in cache to make room.'''
        num, maxcull = 0, self._maxcull
        # Cull number of items allowed (set by self._maxcull)
        for key in self.keys():
            # Remove only maximum # of items allowed by maxcull
            if num <= maxcull:
                # Remove items if expired
                if self.get(key) is None: num += 1
            else: break
        # Remove any additional items up to max # of items allowed by maxcull
        while len(self.keys()) >= self._max_entries and num <= maxcull:
            # Cull remainder of allowed quota at random
            self.delete(random.choice(self.keys()))
            num += 1   


class URLRelay(object):

    '''Passes HTTP requests to a WSGI callable based on URL path component and
    HTTP request method.
    '''

    def __init__(self, **kw):
        # Values can be 'regex' or 'glob'
        pattern_type = kw.get('pattern_type', 'regex')
        # Add any iterable of pairs consisting of a path pattern and either a
        # callback name or a dictionary of HTTP method/callback names
        self._paths = tuple(
            (pattern_compile(u, pattern_type), v) 
            for u, v in kw.get('paths', _reg.get())
        )
        # Shortcut for full module search path
        self._modpath = kw.get('modpath', '')
        # 404 handler
        self._response = kw.get('handler', _handler)
        # Default function
        self._default = kw.get('default')
        # Set maximum number of items to cull from cache if over max
        self._maxcull = kw.get('maxcull', 10)
        # Set cache max entries
        self._max_entries = kw.get('max_entries', 300)
        # Set cache time out
        self._timeout = kw.get('timeout', 300)        

    def __call__(self, env, start_response):
        try:
            # Fetch app and any positional or keyword arguments in path
            app, arg, kw = self.resolve(env['PATH_INFO'], env['REQUEST_METHOD'])
            # Place args in environ dictionary
            env['wsgiorg.routing_args'] = (arg, kw)            
        except (ImportError, AttributeError):
            # Return 404 handler for any exceptions
            return self._response(env, start_response) 
        return app(env, start_response)
    
    @lazy
    def _cache(self):
        '''URL <-> callable mapping Cache.'''
        return MemoryCache(
            timeout=self._timeout, 
            maxcull=self._maxcull, 
            max_entries=self._max_entries,
        )

    def _getapp(self, app):
        '''Loads a callable based on its name
    
        @param app An WSGI application'''
        if isinstance(app, basestring):
            try:
                # Add shortcut to module if present
                dot = app.rindex('.')
                # Import module
                return getattr(__import__(app[:dot], '', '', ['']), app[dot+1:])
            # If nothing but module name, import the module
            except ValueError:
                return __import__(app, '', '', [''])
        return app        
    
    def _loadapp(self, app):
        '''Loads an application based on its name.

        @param app Web application name'''
        # Add module shortcut to module string
        if self._modpath != '' and isinstance(app, basestring):
            app = '.'.join([self._modpath, app])
        newapp = self._getapp(app)
        return newapp

    def resolve(self, path, method):
        '''Fetches a WSGI app based on URL path component and method.

        @param path URL path component
        @param method HTTP method
        '''
        key = ':'.join([path, method])
        # Try fetching app from cache
        app = self._cache.get(key)
        if app is not None: return app
        # Loop through path patterns -> applications
        for pattern, app in self._paths:
            # Test path for match
            search = pattern.search(path)
            # Continue with next iteration if no match
            if not search: continue
            # Get any app specifically linked to an HTTP method
            if isinstance(app, dict): 
                app = app.get(method)
                if app is None: continue             
            app = self._loadapp(app)
            # Ensure we have a callable
            assert hasattr(app, '__call__')
            # Extract any keyword arguments in the path
            kw = search.groupdict()
            # Extract any positional arguments
            args = tuple(i for i in search.groups() if i not in kw)
            # Cache app, positional and keyword arguments
            self._cache[key] = (app, args, kw)
            return app, args, kw
        # Return default app if no matching path and default app is set
        if self._default is not None:
            default = self._loadapp(self._default)
            return default, (), {}
        raise ImportError()


# Copyright (c) 2006 L. C. Rees
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice, 
#       this list of conditions and the following disclaimer.
#    
#    2. Redistributions in binary form must reproduce the above copyright 
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#    3. Neither the name of Django nor the names of its contributors may be used
#       to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Unit tests for urlrelay.'''

import urlrelay
import unittest

def dummy_sr(status, headers, exc_info=None):
    pass

@urlrelay.url(r'^/$')
def index(environ, start_response):
    start_response('200 OK', [('Content-Type', 'text/plain')])
    return ['index']

@urlrelay.url(r'^/handle$', 'GET')
def get_handle(environ, start_response):
    start_response('200 OK', [('Content-Type', 'text/plain')])
    return ['GET']

@urlrelay.url(r'^/handle$', 'POST')
def post_handle(environ, start_response):
    start_response('200 OK', [('Content-Type', 'text/plain')])
    return ['POST']

@urlrelay.url(r'^/handle$', 'PUT')
def put_handle(environ, start_response):
    start_response('200 OK', [('Content-Type', 'text/plain')])
    return ['PUT']    

@urlrelay.url(r'^/argtest/(?P<kwarg1>\w+)/(?P<kwarg2>\w+)/(\w+)/(\w+)$')
def argtests(environ, start_response):
    start_response('200 OK', [('Content-Type', 'text/plain')])
    args, kw = environ['wsgiorg.routing_args']
    return [' '.join([args[0], args[1], kw['kwarg1'], kw['kwarg2']])]

@urlrelay.url(r'^/methtest/(?P<kwarg1>\w+)/(?P<kwarg2>\w+)/(\w+)/(\w+)$', 'GET')
def meth_get(environ, start_response):
    start_response('200 OK', [('Content-Type', 'text/plain')])
    args, kw = environ['wsgiorg.routing_args']
    return [' '.join(['GET', args[0], args[1], kw['kwarg1'], kw['kwarg2']])] 

@urlrelay.url(r'^/methtest/(?P<kwarg1>\w+)/(?P<kwarg2>\w+)/(\w+)/(\w+)$', 'POST')
def meth_post2(environ, start_response):
    start_response('200 OK', [('Content-Type', 'text/plain')])
    args, kw = environ['wsgiorg.routing_args']
    return [' '.join(['POST', args[0], args[1], kw['kwarg1'], kw['kwarg2']])] 

@urlrelay.url(r'^/methtest/(?P<kwarg1>\w+)/(?P<kwarg2>\w+)/(\w+)/(\w+)$', 'PUT')
def meth_put(environ, start_response):
    start_response('200 OK', [('Content-Type', 'text/plain')])
    args, kw = environ['wsgiorg.routing_args']
    return [' '.join(['PUT', args[0], args[1], kw['kwarg1'], kw['kwarg2']])]

@urlrelay.url(r'/spectest/', 'POST')
def meth_post(environ, start_response):
    start_response('200 OK', [('Content-Type', 'text/plain')])
    return ['Empty post']

@urlrelay.url(r'/spectest/(.*)', 'GET')
def meth_get2(environ, start_response):
    start_response('200 OK', [('Content-Type', 'text/plain')])
    args, kw = environ['wsgiorg.routing_args']
    return [' '.join(['GET', args[0]])]

urlrelay.register('^/notfound$', 'urlrelay._handler')
urlrelay.register('^/notfound2$', 'urlrelay._handler', 'GET')
urlrelay.register('^/notfound2$', 'urlrelay._handler', 'PUT')
urlrelay.register('^/notfound2$', 'urlrelay._handler', 'POST')
urlrelay.register('^/notfound4$', '_handler')
urlrelay.register('^/notfound3$', '_handler', 'GET')
urlrelay.register('^/notfound3$', '_handler', 'PUT')
urlrelay.register('^/notfound3$', '_handler', 'POST')

class UrlRelayTest(unittest.TestCase):      

    def test_inmem_root(self):
        '''Checks simple url.'''
        environ = {'PATH_INFO':'/', 'REQUEST_METHOD':'GET'}
        result = urlrelay.URLRelay()(environ, dummy_sr)
        self.assertEqual(result[0], 'index')

    def test_inmem_method_get(self):
        '''Checks url + get method.'''
        environ = {'PATH_INFO':'/handle', 'REQUEST_METHOD':'GET'}
        result = urlrelay.URLRelay()(environ, dummy_sr)
        self.assertEqual(result[0], 'GET')

    def test_inmem_method_post(self):
        '''Checks url + post method.'''
        environ = {'PATH_INFO':'/handle', 'REQUEST_METHOD':'POST'}
        result = urlrelay.URLRelay()(environ, dummy_sr)
        self.assertEqual(result[0], 'POST')        

    def test_inmem_method_put(self):
        '''Checks url + put method.'''
        environ = {'PATH_INFO':'/handle', 'REQUEST_METHOD':'PUT'}
        result = urlrelay.URLRelay()(environ, dummy_sr)
        self.assertEqual(result[0], 'PUT')

    def test_inmem_argextract(self):
        '''Checks arg/kwarg extraction from URL.'''
        environ = {
            'PATH_INFO':'/argtest/kwarg1/kwarg2/arg1/arg2',
            'REQUEST_METHOD':'GET'
        }
        result = urlrelay.URLRelay()(environ, dummy_sr)
        self.assertEqual(result[0], 'arg1 arg2 kwarg1 kwarg2')

    def test_inmem_method_extract_get(self):
        '''Checks arg/kwarg extraction from URL + get method.'''
        environ = {
            'PATH_INFO':'/methtest/kwarg1/kwarg2/arg1/arg2',
            'REQUEST_METHOD':'GET'
        }
        result = urlrelay.URLRelay()(environ, dummy_sr)
        self.assertEqual(result[0], 'GET arg1 arg2 kwarg1 kwarg2')

    def test_inmem_method_extract_post(self):
        '''Checks arg/kwarg extraction from URL + post method.'''
        environ = {
            'PATH_INFO':'/methtest/kwarg1/kwarg2/arg1/arg2',
            'REQUEST_METHOD':'POST'
        }
        result = urlrelay.URLRelay()(environ, dummy_sr)
        self.assertEqual(result[0], 'POST arg1 arg2 kwarg1 kwarg2')

    def test_inmem_method_extract_put(self):
        '''Checks arg/kwarg extraction from URL + put method.'''
        environ = {
            'PATH_INFO':'/methtest/kwarg1/kwarg2/arg1/arg2',
            'REQUEST_METHOD':'PUT'
        }
        result = urlrelay.URLRelay()(environ, dummy_sr)
        self.assertEqual(result[0], 'PUT arg1 arg2 kwarg1 kwarg2')

    def test_offdisk_method(self):
        '''Checks loading handler off of a disk.'''
        environ = {'PATH_INFO':'/notfound', 'REQUEST_METHOD':'GET'}
        result = urlrelay.URLRelay()(environ, dummy_sr)
        self.assertEqual(
            result[0], 
            'Requested URL was not found on this server.'
        )

    def test_offdisk_method_get(self):
        '''Checks loading handler off of a disk + get method.'''
        environ = {'PATH_INFO':'/notfound2', 'REQUEST_METHOD':'GET'}
        result = urlrelay.URLRelay()(environ, dummy_sr)
        self.assertEqual(
            result[0], 
            'Requested URL was not found on this server.'
        )

    def test_offdisk_method_put(self):
        '''Checks loading handler off of a disk + put method.'''
        environ = {'PATH_INFO':'/notfound2', 'REQUEST_METHOD':'PUT'}
        result = urlrelay.URLRelay()(environ, dummy_sr)
        self.assertEqual(
            result[0], 
            'Requested URL was not found on this server.'
        )

    def test_offdisk_method_post(self):
        '''Checks loading handler off of a disk + post method.'''
        environ = {'PATH_INFO':'/notfound2', 'REQUEST_METHOD':'POST'}
        result = urlrelay.URLRelay()(environ, dummy_sr)
        self.assertEqual(
            result[0], 
            'Requested URL was not found on this server.'
        )         

    def test_offdisk_method_modpath(self):
        '''Checks loading handler off of a disk.'''
        environ = {'PATH_INFO':'/notfound4', 'REQUEST_METHOD':'GET'}
        result = urlrelay.URLRelay(modpath='urlrelay')(environ, dummy_sr)
        self.assertEqual(
            result[0], 
            'Requested URL was not found on this server.'
        )

    def test_offdisk_method_get_modpath(self):
        '''Checks loading handler off of a disk + get method.'''
        environ = {'PATH_INFO':'/notfound3', 'REQUEST_METHOD':'GET'}
        result = urlrelay.URLRelay(modpath='urlrelay')(environ, dummy_sr)
        self.assertEqual(
            result[0], 
            'Requested URL was not found on this server.'
        )

    def test_offdisk_method_put_modpath(self):
        '''Checks loading handler off of a disk + put method.'''
        environ = {'PATH_INFO':'/notfound3', 'REQUEST_METHOD':'PUT'}
        result = urlrelay.URLRelay(modpath='urlrelay')(environ, dummy_sr)
        self.assertEqual(
            result[0], 
            'Requested URL was not found on this server.'
        )

    def test_offdisk_method_post_modpath(self):
        '''Checks loading handler off of a disk + post method.'''
        environ = {'PATH_INFO':'/notfound3', 'REQUEST_METHOD':'POST'}
        result = urlrelay.URLRelay(modpath='urlrelay')(environ, dummy_sr)
        self.assertEqual(
            result[0], 
            'Requested URL was not found on this server.'
        )

    def test_inmem_defaultapp(self):
        '''Checks using default app url.'''
        environ = {'PATH_INFO':'/plt', 'REQUEST_METHOD':'GET'}
        result = urlrelay.URLRelay(default=index)(environ, dummy_sr)
        self.assertEqual(result[0], 'index')        

    def test_inmem_defaultapp2(self):
        '''Checks using default app url.'''
        environ = {'PATH_INFO':'/plt', 'REQUEST_METHOD':'GET'}
        result = urlrelay.URLRelay(default=index)(environ, dummy_sr)
        self.assertEqual(result[0], 'index')        

    def test_ondisk_defaultapp(self):
        '''Checks arg/kwarg extraction from URL + put method.'''
        environ = {'PATH_INFO':'/methkwarg1', 'REQUEST_METHOD':'GET'}
        result = urlrelay.URLRelay(
            default='urlrelay._handler'
        )(environ, dummy_sr)
        self.assertEqual(
            result[0], 
            'Requested URL was not found on this server.'
        )

    def test_ondisk_defaultapp_modpath(self):
        '''Checks arg/kwarg extraction from URL + put method.'''
        environ = {'PATH_INFO':'/methkwarg1', 'REQUEST_METHOD':'GET'}
        result = urlrelay.URLRelay(
            default='_handler', 
            modpath='urlrelay'
        )(environ, dummy_sr)
        self.assertEqual(
            result[0], 
            'Requested URL was not found on this server.'
        )

    def test_handler_override(self):
        '''Checks overriding the default 404 handler.'''
        def handle(environ, start_response):                
            start_response('404 Not Found', [('content-type', 'text/plain')])
            return ['404']
        environ = {'PATH_INFO':'/methkwarg1', 'REQUEST_METHOD':'GET'}
        result = urlrelay.URLRelay(handler=handle)(environ, dummy_sr)
        self.assertEqual(result[0], '404')

    def test_notfound(self):
        '''Checks that default 404 handler responds.'''
        environ = {'PATH_INFO':'/methkwarg1', 'REQUEST_METHOD':'GET'}
        result = urlrelay.URLRelay()(environ, dummy_sr)
        self.assertEqual(
            result[0],
            'Requested URL was not found on this server.'
        )

    def test_nonregistry_paths_index(self):
        '''Checks use of non-global path registry.'''
        environ = {'PATH_INFO':'/', 'REQUEST_METHOD':'GET'}        
        tpaths = (
            (r'^/$', index), 
            (r'^/handle$', {
                'GET':get_handle,'POST':post_handle, 'PUT':put_handle
            })
        )
        result = urlrelay.URLRelay(paths=tpaths)(environ, dummy_sr)
        self.assertEqual(result[0], 'index')

    def test_nonregistry_paths_get(self):
        '''Checks use of non-global path registry.'''
        environ = {'PATH_INFO':'/handle', 'REQUEST_METHOD':'GET'}        
        tpaths = (
            (r'^/$', index),
            (r'^/handle$', {
                'GET':get_handle,'POST':post_handle, 'PUT':put_handle
            })
        )
        result = urlrelay.URLRelay(paths=tpaths)(environ, dummy_sr)
        self.assertEqual(result[0], 'GET')

    def test_nonregistry_paths_post(self):
        '''Checks use of non-global path registry.'''
        environ = {'PATH_INFO':'/handle', 'REQUEST_METHOD':'POST'}        
        tpaths = (
            (r'^/$', index), 
            (r'^/handle$', {
                'GET':get_handle,'POST':post_handle, 'PUT':put_handle
            })
        )
        result = urlrelay.URLRelay(paths=tpaths)(environ, dummy_sr)
        self.assertEqual(result[0], 'POST')

    def test_nonregistry_paths_put(self):
        '''Checks use of non-global path registry.'''
        environ = {'PATH_INFO':'/handle', 'REQUEST_METHOD':'PUT'}        
        tpaths = (
            (r'^/$', index),
            (r'^/handle$', {
                'GET':get_handle,'POST':post_handle, 'PUT':put_handle
            })
        )
        result = urlrelay.URLRelay(paths=tpaths)(environ, dummy_sr)
        self.assertEqual(result[0], 'PUT')

    def test_noncallable_in_registry(self):
        '''Checks that non-callables or module strings are not allowed in
        non-global path registry.'''
        environ = {'PATH_INFO':'/', 'REQUEST_METHOD':'PUT'}
        def tempfunc(): 
            test = urlrelay.URLRelay(paths=((r'^/$', []),))
            test(environ, dummy_sr)
        self.assertRaises(AssertionError, tempfunc)
        
    def test_call_more_specific(self):
        '''Checks calling a more-specific URL than one that has a method
        associated with it.'''
        environ = {'PATH_INFO': '/spectest/foo', 'REQUEST_METHOD': 'GET'}
        result = urlrelay.URLRelay()(environ, dummy_sr)
        self.assertEquals('GET foo', result[0])        
    
if __name__ == '__main__': unittest.main()


# Copyright (c) 2006 L. C. Rees
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice, 
#       this list of conditions and the following disclaimer.
#    
#    2. Redistributions in binary form must reproduce the above copyright 
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#    3. Neither the name of Django nor the names of its contributors may be used
#       to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''setup - setuptools based setup for urlrelay.'''

import ez_setup
ez_setup.use_setuptools()

try:
    from setuptools import setup
except:
    from distutils.core import setup

setup(name='urlrelay',
      version='0.7.1',
      description='''RESTful WSGI URL dispatcher.''',
      long_description='''Simple URL dispatcher that passes HTTP
requests to a WSGI application based on a matching URL path regex
pattern and an optional HTTP request method.

Usage example:

#!/bin/env python

import urlrelay

# Simple URL to application mapping
urlrelay.url('^/$')
def index(environ, start_response):
    start_response('200 OK', [('Content-type', 'text/plain')])
    return ['Home Page']
    
# "RESTful" URL to application mapping
urlrelay.url('^/hello_world$', 'GET')
def hello_world(environ, start_response):
    start_response('200 OK', [('Content-type', 'text/plain')])
    return ['Hello World']
    
# URL to on-disk application mapping
# urlrelay.register('^/ondisk$', 'module.on_disk')    

if __name__ == '__main__':
    from wsgiref.simple_server import make_server
    http = make_server('', 8080, urlrelay.URLRelay())
    http.serve_forever()''',
      author='L. C. Rees',
      author_email='lcrees@gmail.com',
      license='BSD',
      py_modules=['urlrelay', 'ez_setup'],
      packages=['tests'],
      test_suite='tests.test_urlrelay',
      zip_safe = True,
      keywords='WSGI URL dispatch relay route middleware web HTTP',
      classifiers=['Development Status :: 4 - Beta',
          'Environment :: Web Environment',
          'License :: OSI Approved :: BSD License',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
          'Topic :: Internet :: WWW/HTTP :: Site Management',
          'Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware'])